#define  XLIB_ILLEGAL_ACCESS
#ifndef lint
static char rcsid[] =
   "$Xukc: xselection.c,v 1.6 91/06/18 15:22:40 rlh2 Exp $";
#endif /* !lint */

/* 
 * Copyright 1990 Richard Hesketh / rlh2@ukc.ac.uk
 *                Computing Lab. University of Kent at Canterbury, UK
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of Richard Hesketh and The University of
 * Kent at Canterbury not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission.
 * Richard Hesketh and The University of Kent at Canterbury make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * Richard Hesketh AND THE UNIVERSITY OF KENT AT CANTERBURY DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL Richard Hesketh OR THE
 * UNIVERSITY OF KENT AT CANTERBURY BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 *
 * Author:  Richard Hesketh / rlh2@ukc.ac.uk, 
 *                Computing Lab. University of Kent at Canterbury, UK
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/Xresource.h>

#define PROG_NAME "xselection"
#define PROG_CLASS "XSelection"

#define CHUNK 1000
#define DEFAULT_PROP_SIZE (10240L)
#define STRLEN(s) (s == NULL ? 0 : strlen(s))

static void send_selection();
static void own_selection();

static Atom selection_atom = (int)NULL;
static unsigned char *current_selection = NULL;
static unsigned char *old_selection = NULL;

#define ClearProp (PropModeReplace + PropModeAppend + PropModePrepend + 1)

static int selection_mode;

static XrmOptionDescRec options[] = {
	{ "-display", ".display", XrmoptionSepArg, (caddr_t)NULL },
	{ "-clear", ".mode", XrmoptionNoArg, (caddr_t)"3" },
	{ "-append", ".mode", XrmoptionNoArg, (caddr_t)"2" },
	{ "-prepend", ".mode", XrmoptionNoArg, (caddr_t)"1" },
	{ "-replace", ".mode", XrmoptionNoArg, (caddr_t)"0" },
	{ "-cutbuffer", ".cutbuffer", XrmoptionNoArg, (caddr_t)"true" },
};

static unsigned char *get_selection();

static void
usage(FILE *output, int return_code)
{
	fprintf(output, "usage: %s -help\n", PROG_NAME);
	fprintf(output, "       %s [ -display display_name ]\n", PROG_NAME);
	fprintf(output, "\t\t  [ -append | -prepend | -replace | -clear ]\n");
	fprintf(output, "\t\t  [ PROPERTY_NAME | -cutbuffer [0-7] ]\n");
	fprintf(output, "\t\t  [ - | [--] new_value ]\n");
	exit (return_code);
}


static char *
emalloc(size)
int size;
{
	/* extern char *malloc(); */
	char *ret;

	if (size == 0)
		size = 1;
	ret = malloc((unsigned)size);

	if (ret == NULL) {
		fprintf(stderr, "%s: malloc failed\n", PROG_NAME);
		exit (-2);
	}
	return (ret);
}


static char *
erealloc(str, size)
char *str;
int size;
{
	/* extern char *realloc(); */
	char *ret;

	if (str == NULL)
		return (emalloc(size));
	if (size == 0)
		size = 1;
	ret = realloc(str, (unsigned)size);

	if (ret == NULL) {
		fprintf(stderr, "%s: realloc failed\n", PROG_NAME);
		exit (-2);
	}
	return (ret);
}


static unsigned char *
get_value_from_stdin()
{
	unsigned char *str;
	int count = 0, num_chunks = 1;
	int ch;

	str = (unsigned char *)emalloc(CHUNK);
	*str = '\0';

	ch = getchar();
	while (ch != EOF) {
		if (count % CHUNK == (CHUNK-3))
			str = (unsigned char *)erealloc(str,
						(++num_chunks * CHUNK));
		str[count++] = ch;
		ch = getchar();
	}
	str[count] = '\0';
	return (str);
}


int
main(argc, argv)
int argc;
char *argv[];
{
	Display *dpy;
	XEvent ev;
	char resource[100];
	char Resource[100];	/* for class information */
	char *blank, *display;
	XrmDatabase odb = NULL, db = NULL;
	XrmValue value;
	Window window;
	int want_to_own = 0;
	Time event_time;

	XrmParseCommand(&odb, options, sizeof(options)/sizeof(options[0]),
			PROG_NAME, &argc, argv);

	(void)sprintf(resource, "%s.display", PROG_NAME);
	(void)sprintf(Resource, "%s.Display", PROG_CLASS);
	if (XrmGetResource(odb, resource, Resource, &blank, &value))
		display = (char *)value.addr;
	else
		display = NULL;

	if ((dpy = XOpenDisplay(display)) == NULL) {
		fprintf(stderr, "%s: Couldn't open display %s\n",
				PROG_NAME, XDisplayName(display));
		exit (-1);
	}

	if ((argc == 2) && (!strcmp(argv[1],"-help")))
		usage(stdout, 0);

	if (argc < 2 || argc > 4)
		usage(stderr, -1);

	XrmMergeDatabases(dpy->db, &db);
	XrmMergeDatabases(odb, &db);

	(void)sprintf(resource, "%s.mode", PROG_NAME);
	(void)sprintf(Resource, "%s.Mode", PROG_CLASS);
	if (XrmGetResource(db, resource, Resource, &blank, &value) ||
	    XrmGetResource(dpy->db, resource, Resource, &blank, &value))
	    switch (value.addr[0]) {
	    case '0':
	            selection_mode = PropModeReplace;
	            break;
	    case '1':
	            selection_mode = PropModePrepend;
	            break;
	    case '2':
	            selection_mode = PropModeAppend;
	            break;
	    case '3':
		    selection_mode = ClearProp;
	    }
	else
	    selection_mode = PropModeReplace;

	if (selection_mode != PropModeReplace &&
	    selection_mode != PropModeAppend &&
	    selection_mode != PropModePrepend &&
	    selection_mode != ClearProp) {
	        fprintf(stderr, "%s: Bad selection mode %d.\n",
	                PROG_NAME, selection_mode);
	        exit(-2);
	}

	if (argc == 3) {
		/* we want to set a property .. find the new value */
		if (strcmp(argv[2], "-") == 0)
			current_selection = get_value_from_stdin();
		else
			current_selection = (unsigned char *)argv[2];
		want_to_own = 1;
	} else if (argc == 4) {
		/* set a property to a value that is the same as a flag */
		if (strcmp(argv[2], "--") == 0)
			/* I bet this will never get used! .. but be safe */
			current_selection = (unsigned char *)argv[3];
		else
			usage(stderr, -1);
		want_to_own = 1;
	}

	(void)sprintf(resource, "%s.cutbuffer", PROG_NAME);
	(void)sprintf(Resource, "%s.Cutbuffer", PROG_CLASS);
	if (XrmGetResource(db, resource, Resource, &blank, &value) ||
	    XrmGetResource(dpy->db, resource, Resource, &blank, &value)) {
		int cutbuffer = atoi(argv[1]);
		if (cutbuffer < 0 || cutbuffer > 7) {
			fprintf(stderr,
			    "%s: %d is not a valid cut buffer, must be 0-7.\n",
			    PROG_NAME, cutbuffer);
			exit(-8);
		}
		if (selection_mode == ClearProp) {
			want_to_own = True;
			selection_mode = PropModeReplace;
			current_selection = (unsigned char *)"";
		}
		if (want_to_own)
			XChangeProperty(dpy, DefaultRootWindow(dpy),
					XA_CUT_BUFFER0 + cutbuffer, XA_STRING,
					8, selection_mode, current_selection,
					strlen(current_selection));
		else {
			Atom actual_type;
			int actual_format;
			unsigned long nitems = 0, bytes_after = 0;
			unsigned char* data = NULL;

			do {
				if (XGetWindowProperty(dpy,
					   DefaultRootWindow(dpy),
					   XA_CUT_BUFFER0 + cutbuffer,
					   nitems/4, DEFAULT_PROP_SIZE, False,
					   AnyPropertyType,
					   &actual_type, &actual_format,
					   &nitems, &bytes_after, &data)
				    == Success) {
					printf("%s", data);
					XFree(data);
				}
			} while (bytes_after > 0);
		}
		XCloseDisplay(dpy);
		exit(0);
	}

	selection_atom = XInternAtom(dpy, argv[1], argc == 2);

	if (selection_atom == (int)NULL) {
		fprintf(stderr,
			"%s: %s is not a name of a known selection property.\n",
			PROG_NAME, argv[1]);
		exit (-2);
	}

	/* this window never gets mapped - its simply used as an ID */
	window = XCreateSimpleWindow(dpy, DefaultRootWindow(dpy),
							0, 0, 1, 1, 0, 0, 0);

#ifdef UNDER_DEVELOPMENT
	/* fake a button event so we can get a timestamp from the server
	 * to use in our XSetSelectionOwner() call.
	 */
	XSelectInput(dpy, window, ButtonPressMask);
	ev.xbutton.type = ButtonPress;
	ev.xbutton.display = dpy;
	ev.xbutton.window = ev.xbutton.subwindow = window;
	ev.xbutton.root = DefaultRootWindow(dpy);
	ev.xbutton.time = CurrentTime;
	if (!XSendEvent(dpy, window, False, ButtonPressMask, &ev)) {
		fprintf(stderr, "%s: failed to not get timestamp\n",
			PROG_NAME);
		exit(1);
	}

	for (;;) {
		XNextEvent(dpy, &ev);
		if (ev.type == ButtonPress &&
		    ev.xbutton.window == window)
			break;
	};
	event_time = ev.xbutton.time;
#else
	event_time = CurrentTime;	/* this is wrong! */
#endif /* UNDER_DEVELOPMENT */

	if (selection_mode == ClearProp || want_to_own) {
		/* set the selection and wait for someone to take it */
		if (selection_mode != PropModeReplace)
			old_selection = get_selection(dpy, window,
						      selection_atom);
		own_selection(dpy, window, selection_atom, event_time);
	} else {
		/* get the selection */
		current_selection = get_selection(dpy, window, selection_atom);
		if (current_selection != NULL)
			printf("%s", current_selection);
		else {
			fprintf(stderr, "%s: failed to get the %s selection.\n",
				PROG_NAME, argv[1]);
			exit(1);
		}
		XCloseDisplay(dpy);
		exit(0);
	}

	if (selection_mode == ClearProp)
		exit(0);

	switch (fork()) {
		case -1:
			fprintf(stderr, "%s: fork failed\n", PROG_NAME);
			break;
		case 0:
			/* child */
			XSelectInput(dpy, window, 0l);

			/* wait here 'til we lose the selection.
			 * respond to any SelectionRequests
			 */
			for (;;) {
				XNextEvent(dpy, &ev);
				if (ev.type == SelectionClear &&
				    ev.xselectionclear.selection
				    == selection_atom)
					break;
				if (ev.type == SelectionRequest &&
				    ev.xselectionrequest.selection
				    == selection_atom)
					send_selection(dpy, &ev);
			};
			XCloseDisplay(dpy);
			exit(0);
			break; /* NOTREACHED */
	}
	return 0;
}


/* Send a copy of the our owned selection to the requestor.
 * We do this by changing the contents of the property given in the
 * SelectionRequest event.
 */
static void
send_selection(dpy, ptr_event)
Display *dpy;
XEvent *ptr_event;
{
        XSelectionEvent notify_event;
        XSelectionRequestEvent  *req_event;
        unsigned char *data;

#ifndef XA_LENGTH
	unsigned long XA_LENGTH = XInternAtom(dpy, "LENGTH", True);
#endif /* !XA_LENGTH */
	Atom XA_TARGETS = XInternAtom(dpy, "TARGETS", True);

	if (ptr_event->xselectionrequest.target == XA_LENGTH) {
		unsigned int length;

		/* tell the requestor how long the selection is */
		length = (selection_mode != PropModeReplace)
			 ? STRLEN(old_selection) + STRLEN(current_selection)
			 : STRLEN(current_selection);

		XChangeProperty(ptr_event->xselectionrequest.display,
				ptr_event->xselectionrequest.requestor,
				ptr_event->xselectionrequest.property,
				ptr_event->xselectionrequest.target,
				32, 
				PropModeReplace, 
				(void*)&length,
				(int)sizeof(unsigned int));

		req_event = &(ptr_event->xselectionrequest);

		notify_event.type      = SelectionNotify;
		notify_event.display   = req_event->display;
		notify_event.requestor = req_event->requestor;
		notify_event.selection = req_event->selection;
		notify_event.target    = req_event->target;
		notify_event.time      = req_event->time;

		if (req_event->property == None)
			notify_event.property = req_event->target;
		else
			notify_event.property = req_event->property;

		(void)XSendEvent(req_event->display, req_event->requestor,
					False, 0, (XEvent *)&notify_event);
		XFlush(dpy);
		return;
	}

	if (ptr_event->xselectionrequest.target == XA_TARGETS) {

		XChangeProperty(ptr_event->xselectionrequest.display,
				ptr_event->xselectionrequest.requestor,
				ptr_event->xselectionrequest.property,
				ptr_event->xselectionrequest.target,
				32, 
				PropModeAppend, 
				NULL,
				0);

		req_event = &(ptr_event->xselectionrequest);

		notify_event.type      = SelectionNotify;
		notify_event.display   = req_event->display;
		notify_event.requestor = req_event->requestor;
		notify_event.selection = req_event->selection;
		notify_event.target    = req_event->target;
		notify_event.time      = req_event->time;

		if (req_event->property == None)
			notify_event.property = req_event->target;
		else
			notify_event.property = req_event->property;

		(void)XSendEvent(req_event->display, req_event->requestor,
					False, 0, (XEvent *)&notify_event);
		XFlush(dpy);
		return;
	}

	if (selection_mode != PropModeReplace) {
		data = old_selection;
		XChangeProperty(ptr_event->xselectionrequest.display,
				ptr_event->xselectionrequest.requestor,
				ptr_event->xselectionrequest.property,
				ptr_event->xselectionrequest.target,
				8, PropModeReplace, data,
				STRLEN(old_selection));
	}

        data = current_selection;

        XChangeProperty(ptr_event->xselectionrequest.display,
                        ptr_event->xselectionrequest.requestor,
                        ptr_event->xselectionrequest.property,
                        ptr_event->xselectionrequest.target,
                        8, selection_mode, data,
			STRLEN(current_selection));

        req_event = &(ptr_event->xselectionrequest);

        notify_event.type      = SelectionNotify;
        notify_event.display   = req_event->display;
        notify_event.requestor = req_event->requestor;
        notify_event.selection = req_event->selection;
        notify_event.target    = req_event->target;
        notify_event.time      = req_event->time;

        if (req_event->property == None)
                notify_event.property = req_event->target;
        else
                notify_event.property = req_event->property;

	/* tell the requestor he has a copy of the selection */
        (void)XSendEvent(req_event->display, req_event->requestor,
					False, 0, (XEvent *)&notify_event);
	XFlush(dpy);
}


/* Request and retrieve a copy of the current selection property */
static unsigned char *
get_selection(dpy, window, atom)
Display *dpy;
Window window;
Atom atom;
{
	XEvent  event;
        static struct   timeval timeout = {3, 0};
        int     server_fd_mask;
        extern int      errno;
        Atom    actual_type;
        int     actual_format;
        unsigned long   nitems, bytes_after;
        unsigned char    *data;
        int     old_inmode;
	unsigned char *already_got;
	unsigned long size;

	/* tell the server we want a copy of the selection */
        XConvertSelection(dpy, atom, XA_STRING, XA_STRING, window, CurrentTime);
        server_fd_mask = (1 << ConnectionNumber(dpy));

        do {
                int     result;

                do {
                        if (XPending(dpy) > 0)
                                result = 1;
                        else {
                                result = select(32, (void*)&server_fd_mask,
						(fd_set  *)NULL, (fd_set  *)NULL,
						&timeout);
                                if (result == -1 && errno != EINTR) {
                                        fprintf(stderr, "%s: select failed in get_selection (errno:%d)\n",
							PROG_NAME, errno);
					exit (-1);
                                }
                        }
                } while (result == -1);

                if (result == 0)
                        return NULL;

                XNextEvent(dpy, &event);
                if (event.type == SelectionRequest)
			send_selection(dpy, window, atom, &event);
        } while (event.type != SelectionNotify);

        if (event.xselection.property == None)
		/* retrieving the selection failed for some reason */
                return NULL;

        if (XGetWindowProperty(dpy,
                event.xselection.requestor,
                event.xselection.property,
                0L, DEFAULT_PROP_SIZE,
                True,
                AnyPropertyType,
                &actual_type,
                &actual_format,
                &nitems,
                &bytes_after,
                (unsigned char **)&data) != Success)
			/* the get failed for some reason */
			return NULL;

	if (bytes_after > 0) {
		already_got = (unsigned char *)erealloc(
					data, nitems + bytes_after + 1);
		size = nitems;
	} else
		return data;

	while (bytes_after > 0) {
		if (XGetWindowProperty(dpy,
			event.xselection.requestor,
			event.xselection.property,
			size/4, DEFAULT_PROP_SIZE,
			True,
			AnyPropertyType,
			&actual_type,
			&actual_format,
			&nitems,
			&bytes_after,
			(unsigned char **)&data) != Success)
				/* the get failed for some reason */
				return already_got;
		else {
			(void)memcpy(already_got + size, data, nitems);
			size += nitems;
			XFree(data);
		}
	}
		
	return already_got;
}


/* tell the server we want to own the selection */
static void
own_selection(dpy, window, atom, event_time)
Display *dpy;
Window window;
Atom atom;
Time event_time;
{
        XSetSelectionOwner(dpy, atom, window, event_time);
	XFlush(dpy);
}
